from collections import Counter, OrderedDict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter


def display_features(features, feature_names, row_names=None):
    """helper function to print values in a table"""
    df = pd.DataFrame(data=features,
                      index=row_names,
                      columns=feature_names)
    print(df)


def bins_for_discrete_hist(data):
    u = np.unique(data)
    
    if len(u) > 1:
        d = np.diff(np.sort(u)).min()
        l = data.min() - float(d)/2
        r = data.max() + float(d)/2 + d
        return np.arange(l, r, d)
    else:
        return 1

     
def plot_discrete_hist(data, **kwargs):
    assert len(data) > 0
    
    if type(data) is not np.ndarray:
        data = np.array(data)
    
    bins = bins_for_discrete_hist(data)
        
    ax = plt.gca()
    ax.set_xticks(np.arange(data.min(), data.max()+1, 1))
    ax.hist(data, bins, **kwargs)
    
    plt.show()

    
def plot_discrete_joint_distrib(x, y, figsize=(8, 8), rwidth=1, xlabel="x", ylabel="y",
                                scatter_kwargs=None, hist_kwargs=None):
    if type(x) is not np.ndarray:
        x = np.array(x)
    if type(y) is not np.ndarray:
        y = np.array(y)
        
    scatter_kwargs = scatter_kwargs or {}
    hist_kwargs = hist_kwargs or {}
    
    # definitions for the axes
    left, width = 0.1, 0.65
    bottom, height = 0.1, 0.65
    bottom_h = left_h = left + width + 0.02

    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom_h, width, 0.2]
    rect_histy = [left_h, bottom, 0.2, height]

    # start with a rectangular Figure
    plt.figure(figsize=figsize)

    # create the axes components
    ax_scatter = plt.axes(rect_scatter)
    ax_hist_x = plt.axes(rect_histx)
    ax_hist_y = plt.axes(rect_histy)

    # no labels for histograms
    nullfmt = NullFormatter()         # no labels
    ax_hist_x.xaxis.set_major_formatter(nullfmt)
    ax_hist_y.yaxis.set_major_formatter(nullfmt)

    # draw scatter plot
    ax_scatter.scatter(x, y, **scatter_kwargs)
    
    # get bins for discrete data
    bins_x = bins_for_discrete_hist(x)
    bins_y = bins_for_discrete_hist(y)
    
    # set limits
    xlim = bins_x[0] - 0.25, bins_x[-1] + 0.25
    ylim = bins_y[0] - 0.25, bins_y[-1] + 0.25

    ax_scatter.set_xlim(xlim)
    ax_scatter.set_ylim(ylim)
    
    # customize axes
    if xlabel:
        ax_scatter.set_xlabel(xlabel)
    if ylabel:
        ax_scatter.set_ylabel(ylabel)
    ax_scatter.set_xticks(np.arange(x.min(), x.max()+1, 1))
    ax_scatter.set_yticks(np.arange(y.min(), y.max()+1, 1))
    
    # draw histograms
    ax_hist_x.hist(x, bins=bins_x, rwidth=rwidth, **hist_kwargs)
    ax_hist_y.hist(y, bins=bins_y, rwidth=rwidth, orientation='horizontal', **hist_kwargs)

    plt.show()


def plot_heatmap(data, xticklabels=None, yticklabels=None, xlabel=None, ylabel=None,
                 title=None, grid=True, values_in_cells=True, round_values_in_cells=2, legend=False,
                 save_to=None):
    """"helper function to plot a heatmap"""
    if type(data) is not np.ndarray:
        data = np.array(data)
    
    fig = plt.figure()
    plt.matshow(data)
    if legend:
        plt.colorbar()
    
    # set title
    if title:
        plt.title(title, y=1.25)
    
    plt.xticks(rotation=45)
    
    n_rows, n_cols = data.shape

    # get Axes object
    ax = plt.gca()

    # draw values in cells
    if values_in_cells:
        textcol_thresh = data.min() + (data.max() - data.min()) / 2
        x_indices, y_indices = np.meshgrid(np.arange(n_cols), np.arange(n_rows))
        for x, y in zip(x_indices.flatten(), y_indices.flatten()):
            val = data[y, x]
            # lower values get white text color for better visibility
            textcol = 'white' if val < textcol_thresh else 'black'
            disp_val = round(val, round_values_in_cells) if round_values_in_cells is not None else val
            ax.text(x, y, disp_val, va='center', ha='center', color=textcol)

    # customize axes
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    ax.set_xticks(np.arange(0, n_cols))
    ax.set_yticks(np.arange(0, n_rows))
    
    if xticklabels:
        ax.set_xticklabels(xticklabels)
    if yticklabels:
        ax.set_yticklabels(yticklabels)

    # gridlines based on minor ticks
    if grid:
        ax.set_xticks(np.arange(-.5, n_cols), minor=True);
        ax.set_yticks(np.arange(-.5, n_rows), minor=True);
        ax.grid(which='minor', color='w', linestyle='-', linewidth=1)

    if save_to:
        plt.savefig(save_to)

    plt.show()


def create_ngrams(tokens, n, join=True, join_str=' '):
    assert 1 < n <= len(tokens)
    ngrams = [[tokens[i+j] for j in range(n)]
              for i in range(len(tokens)-n+1)]
    if join:
        return list(map(lambda x: join_str.join(x), ngrams))
    else:
        return ngrams


def create_bow(documents):
    if type(documents) is dict:
        counts = sorted([(doc_label, Counter(tok)) for doc_label, tok in documents.items()], key=lambda x: x[0])
    elif type(documents) in (list, tuple):
        counts = [(str(i), Counter(tok)) for i, tok in enumerate(documents)]
    else:
        raise ValueError('`documents` must be a dict, list or tuple.')
    
    counts = OrderedDict(counts)
    
    vocab = set()
    for counter in counts.values():
        vocab |= set(counter.keys())   # set union of unique tokens per document

    vocab = sorted(list(vocab))  # sorting here only for better display later

    bow = []
    for counter in counts.values():  # iterate through each document counter instance
        # make a list that contains the term count of each term in this document
        # if a term of the vocab. does not exist in this document, set it to 0 (default value of .get())
        bow_row = [counter.get(term, 0) for term in vocab]
        bow.append(bow_row)

    doc_labels = list(counts.keys())
    
    return bow, doc_labels, vocab

def create_repeated_indices_from_dtm(dtm):
    d_ind, w_ind = np.nonzero(dtm)   # indices for non-zero occurrences of words
    # d_ind .. document index for kth unique word
    # w_ind .. word index for the kth unique word
    
    counts = dtm[d_ind, w_ind]   # non-zero word counts as array
    n_total_words = np.sum(counts)
    
    di = np.repeat(d_ind, counts)   # repeat doc. indices according to word counts
    
    wi = np.empty(n_total_words, dtype=int)   # repeat word indices according to word counts
    fill_idx = 0
    for i, n in enumerate(counts):
        wi[fill_idx:fill_idx+n] = w_ind[i]
        fill_idx += n

    assert len(di) == len(wi)
    
    return di, wi


def labelled_top_n_from_distribution(distrib, labels, n, join_str=None):
    assert len(distrib) > 0
    assert n > 0
    top_ind = np.argsort(distrib)[::-1][:n]
    if type(labels) is not np.ndarray:
        labels = np.array(labels)
    
    top_labels = labels[top_ind]
    
    if join_str is not None:
        return join_str.join(top_labels)
    else:
        return top_labels